import numpy as np
from scipy.sparse import diags
from scipy.sparse.linalg import spsolve
import matplotlib.pyplot as plt

#========================================================
# numerical calculation of a laminar boundary layer 
# velocity profile using a finite-volume approach
#========================================================
#
#      0     1     2            N-1    N    N+1
#   ------------------- - - - -------------------
#   |     |     |     |       |     |     |     |
#   |  o  |  o  |  o  |       |  o  |  o  |  o  |
#   | G1  |     |     |       |     |     | G2  |
#   ------------------- - - - -------------------
#         ^                               ^      
#      	 y=0.                           y=h/2    
#        u=0.                          du/dy=0
#
# The computational domain contains N cells ranging 
# from cell 1 to cell N. G1 and G2 are ghost cells 
# (cells outside of the computational domain used for
# boundary condition implementation).
#========================================================

if __name__ == "__main__":
	
	savefig = 1

	h     =  0.06   # channel height
	mu    =  1.8e-5 # viscosity
	rho   =  1.2    # density
	dpdx  = -0.01   # pressure gradient

	# number of cells

	N = 10 # <= CHANGE CELL COUNT HERE

	# allocate vectors

	y = np.zeros(N+2)
	U = np.zeros(N+2)

	# calculate cell size (dy) assuming equidistant grid

	dy=(h/2.0)/N

	# add wall-normal coordinates for all cells (cell 1 to N) 

	y[slice(1,N+1)] = np.linspace(dy/2.0,0.5*h-(dy/2.0),N)

	# add centerline coordinate

	y[-1]=h/2.

	# calculate the velocity for all y-coordinates (analytic solution)

	Ua = ... # <= ADD ANALYTIC SOLUTION HERE Ua=f(y)

	# generate matrix diagonal vector and off-diagonal vectors

	diagonal  = -2.0*np.ones(N)
	off_diagonal = np.ones(N-1)

	# set boundary conditions (update diagonal elements)

	diagonal[ 0] = ... # <= ADD LEFT-END BOUNDARY CONDITION HERE
	diagonal[-1] = ... # <= ADD RIGHT-END BOUNDARY CONDITION HERE

	# generate sparse matrix (A)

	A = diags([diagonal,off_diagonal,off_diagonal],[0, -1, 1],format="csr")

	# generate right-hand-side vector (b)

	b = (dpdx*dy**2/mu)*np.ones(N)

	# solve linear equation system AU=b => U=inv(A)b

	U[slice(1,N+1)] = spsolve(A,b)

	# update boundary velocity

	U[-1] = U[-2]

	# plot velocity profile

	fig = plt.figure()
	ax  = fig.add_subplot(111)
	ax.plot(U,y,linewidth=2,label='numeric')
	ax.plot(Ua,y,linewidth=2,label='analytic')
	ax.set_xlabel('axial velocity (u [m/s])',fontsize=14)
	ax.set_ylabel('wall-normal coordinate (y [m])',fontsize=14)
	ax.set_title('Velocity profile - laminar channel flow',fontsize=14)
	ax.legend(fontsize=14)
	ax.set_xlim(0,0.30)
	ax.set_ylim(0,0.5*h)
	for label in (ax.get_xticklabels() + ax.get_yticklabels()):
		label.set_fontsize(14)
	ax.grid()
	if savefig:
		plt.savefig('laminar-profile.png',bbox_inches='tight')
	else:
		plt.show()

